//
// ********************************************************************
// * License and Disclaimer                                           *
// *                                                                  *
// * The  Geant4 software  is  copyright of the Copyright Holders  of *
// * the Geant4 Collaboration.  It is provided  under  the terms  and *
// * conditions of the Geant4 Software License,  included in the file *
// * LICENSE and available at  http://cern.ch/geant4/license .  These *
// * include a list of copyright holders.                             *
// *                                                                  *
// * Neither the authors of this software system, nor their employing *
// * institutes,nor the agencies providing financial support for this *
// * work  make  any representation or  warranty, express or implied, *
// * regarding  this  software system or assume any liability for its *
// * use.  Please see the license in the file  LICENSE  and URL above *
// * for the full disclaimer and the limitation of liability.         *
// *                                                                  *
// * This  code  implementation is the result of  the  scientific and *
// * technical work of the GEANT4 collaboration.                      *
// * By using,  copying,  modifying or  distributing the software (or *
// * any work based  on the software)  you  agree  to acknowledge its *
// * use  in  resulting  scientific  publications,  and indicate your *
// * acceptance of all terms of the Geant4 Software license.          *
// ********************************************************************
//

#include "VKDetectorConstruction.hh"
#include "VKRecorderRoot.hh"
#include "VKBetaDetector.hh"
#include "VKBetaDetectorV2.hh"
#include "VKRecorderRootMessenger.hh"
#include "VKPrimaryGeneratorAction.hh"
#include "VKUserEventInformation.hh"
#include "VKScintillatorHit.hh"
#include "VKSiPMHit.hh"

#include "G4VPhysicalVolume.hh"
#include "G4RunManager.hh"
#include "G4EventManager.hh"

#include <TNamed.h>
#include <TParameter.h>
#include <TObjString.h>

#include <sstream> // needed by rdbuf()

using namespace std;


VKRecorderRoot::VKRecorderRoot(G4String name)
 : fFilename(name),
   fZeroSuppression(false),
   fRunMacroFileName("")
{
    fTFile = NULL;
    fDetTree = fTraTree = NULL;

    fFocalPlane = fMidwayPlane = NULL;
    fSignalPV = fVetoPV = NULL;
    fFrontV2PV = fSignalV2PV = fVetoV2PV = NULL;

    // create messenger class
    fRecorderRoot_msg = new VKRecorderRootMessenger(this);
}


VKRecorderRoot::~VKRecorderRoot() 
{
    // make sure ROOT file has been closed
    if (fTFile != NULL && fTFile->IsOpen()) CloseRootFile();
    
    // delete pointers 
    delete fTFile;
    delete fDetTree;
    delete fTraTree;
}


void VKRecorderRoot::RecordBeginOfRun(const G4Run*)
{
    // detector
    fDetector = static_cast<const VKDetectorConstruction*>
        (G4RunManager::GetRunManager()->GetUserDetectorConstruction());
        
    // which detectors are in the setup?
    fBetaOn    = fDetector -> GetBetaDetectorOn();
    fBetaV2On  = fDetector -> GetBetaDetectorV2On();
    fVirtualOn = fDetector -> GetVirtualDetectorsOn();
        
    // get pointers to beta-detector v1.0 phys volumes
    if (fBetaOn) {
        fSignalPV = fDetector->GetBetaDetector()->GetSignalPhysVol();
        fVetoPV = fDetector->GetBetaDetector()->GetVetoPhysVol();
    }

    // get pointers to beta-detector v2.0 phys volumes
    if (fBetaV2On) {
        fFrontV2PV = fDetector->GetBetaDetectorV2()->GetFrontPhysVol();
        fSignalV2PV = fDetector->GetBetaDetectorV2()->GetSignalPhysVol();
        fVetoV2PV = fDetector->GetBetaDetectorV2()->GetVetoPhysVol();
    }
        
    // get pointers to virtual detectors
    if (fVirtualOn) {
        fFocalPlane = fDetector->GetFocalPlane();
        fMidwayPlane = fDetector->GetMidwayPlane();
    }    

    // Only create TFile and trees if filename is provided
    if (fFilename != "") OpenRootFile(fFilename);
    
    // dump settings to Root file
    WriteMetaDataToRootFile();
}


void VKRecorderRoot::RecordEndOfRun(const G4Run*)
{
    WriteToRootFileAndClose();
}


void VKRecorderRoot::RecordBeginOfEvent(const G4Event* anEvent)
{
    // clear variables
    fX_fp = fY_fp = fZ_fp = fEkin_fp = fTheta_fp = fPhi_fp = 0;
    fX_mp = fY_mp = fZ_mp = fEkin_mp = fTheta_mp = fPhi_mp = 0;
    fX_so = fY_so = fZ_so = fEkin_so = fTheta_so = fPhi_so = 0;

    fEDepSignal = 0;
    fEDepVeto   = 0;

    fEDepV2Front  = 0;
    fEDepV2Signal = 0;
    fEDepV2Veto   = 0;

    fPhotonsSignal = 0;
    fPhotonsVeto   = 0;
    
    // clear Root branches
    ClearRootBranches();
    
    // Record data about particle source:
    
    // vertex
    fPrimVertex = anEvent -> GetPrimaryVertex(0); 
        // NB: if several vertices exist, this will only get the first one

    // particle
    fPrimParticle = fPrimVertex -> GetPrimary(0);   
        // NB: if several primary particles exist, this will only get the first one

    // position
    G4ThreeVector pos = fPrimVertex -> GetPosition();
    
    // momentum and mass
    G4ThreeVector mom = fPrimParticle -> GetMomentum(); 
    G4double mass = fPrimParticle -> GetMass(); 
    
    // store data
    fX_so     = pos.x();
    fY_so     = pos.y();
    fZ_so     = pos.z();
    fEkin_so  = pow(mom.mag2()+pow(mass,2), 0.5) - mass;
    fTheta_so = mom.getTheta();
    fPhi_so   = mom.getPhi();
}


void VKRecorderRoot::RecordEndOfEvent(const G4Event* anEvent)
{
    // event info
    VKUserEventInformation* eventInfo
        = (VKUserEventInformation*)anEvent -> GetUserInformation();

    int id = -1;

    // deposited energy
    if (fBetaOn) {
        id++;
        fEDepSignal    = eventInfo->GetEDep(id);
        fPhotonsSignal = eventInfo->GetHitCount(id);
        id++;
        fEDepVeto      = eventInfo->GetEDep(id);
        fPhotonsVeto   = eventInfo->GetHitCount(id);
    }
    
    // deposited energy (v2.0)        
    if (fBetaV2On) {
        id++;
        fEDepV2Front  = eventInfo->GetEDep(id);
        id++;
        fEDepV2Signal = eventInfo->GetEDep(id);
        id++;
        fEDepV2Veto   = eventInfo->GetEDep(id);
    }
        
    // fill ROOT tree
    FillTree();
}


void VKRecorderRoot::RecordTrack(const G4Track*)
{}


void VKRecorderRoot::RecordStep(const G4Step* aStep)
{
    // pre-step physical volume
    G4VPhysicalVolume *prePV = aStep -> GetPreStepPoint() -> GetPhysicalVolume();

    // next physical volume
    G4VPhysicalVolume *nextPV = aStep -> GetTrack() -> GetNextVolume();
    
    // process type
    G4String procName = aStep -> GetPostStepPoint() -> GetProcessDefinedStep() -> GetProcessName();

    // particle type
    G4String partType = aStep -> GetTrack() -> GetDynamicParticle() -> 
                                                    GetDefinition() -> GetParticleName();

    // --------------------------- //
    //     VIRTUAL DETECTORS       //
    // --------------------------- //

    // If setup contains virtual detectors, process is transportation and particle is electron
    if (fVirtualOn && procName == "Transportation" && partType == "e-") {
        
        // Is electron entering focal plane or midway plane? 
        G4bool enterFP = (prePV != fFocalPlane && nextPV == fFocalPlane);
        G4bool enterMP = (prePV != fMidwayPlane && nextPV == fMidwayPlane);
        if (enterFP || enterMP) {
        
            // get kinematic data
            G4ThreeVector pos = aStep->GetTrack()->GetPosition();
            G4double ekin     = aStep->GetTrack()->GetKineticEnergy();
            G4ThreeVector dir = aStep->GetTrack()->GetMomentumDirection();

            if (enterFP) {
                fX_fp     = pos.x();
                fY_fp     = pos.y();
                fZ_fp     = pos.z();
                fEkin_fp  = ekin;
                fTheta_fp = dir.getTheta();
                fPhi_fp   = dir.getPhi();
            }
            else {
                fX_mp     = pos.x();
                fY_mp     = pos.y();
                fZ_mp     = pos.z();
                fEkin_mp  = ekin;
                fTheta_mp = dir.getTheta();
                fPhi_mp   = dir.getPhi();            
            }
        }
    }
}


// detector branches
void VKRecorderRoot::CreateDetBranches() 
{
    if (fTFile == NULL) return;

    if (fBetaOn) {
        fDetTree -> Branch("EDepSignal", &_eDepSignal, "_eDepSignal/F");
        fDetTree -> Branch("EDepVeto"  , &_eDepVeto,   "_eDepVeto/F");

        fDetTree -> Branch("PhotonsSignal", &_photonsSignal, "_photonsSignal/i");
        fDetTree -> Branch("PhotonsVeto"  , &_photonsVeto  , "_photonsVeto/i");
    }
    
    if (fBetaV2On) {
        fDetTree -> Branch("EDepV2Front",  &_eDepV2Front,  "_eDepV2Front/F");
        fDetTree -> Branch("EDepV2Signal", &_eDepV2Signal, "_eDepV2Signal/F");
        fDetTree -> Branch("EDepV2Veto"  , &_eDepV2Veto,   "_eDepV2Veto/F");
    }
}


// trajectory branches
void VKRecorderRoot::CreateTraBranches() 
{
    if (fTFile == NULL) return;

    // focal plane
    fTraTree -> Branch("FocalX",   &_Xfp,   "_Xfp/F");
    fTraTree -> Branch("FocalY",   &_Yfp,   "_Yfp/F");
    fTraTree -> Branch("FocalZ",   &_Zfp,   "_Zfp/F");
    fTraTree -> Branch("FocalE",   &_Efp,   "_Efp/F");
    fTraTree -> Branch("FocalTH",  &_THfp,  "_THfp/F");
    fTraTree -> Branch("FocalPHI", &_PHIfp, "_PHIfp/F");
    // midway plane
    fTraTree -> Branch("MidwayX",   &_Xmp,   "_Xmp/F");
    fTraTree -> Branch("MidwayY",   &_Ymp,   "_Ymp/F");
    fTraTree -> Branch("MidwayZ",   &_Zmp,   "_Zmp/F");
    fTraTree -> Branch("MidwayE",   &_Emp,   "_Emp/F");
    fTraTree -> Branch("MidwayTH",  &_THmp,  "_THmp/F");
    fTraTree -> Branch("MidwayPHI", &_PHImp, "_PHImp/F");
    // source
    fTraTree -> Branch("SourceX",   &_Xso,   "_Xso/F");
    fTraTree -> Branch("SourceY",   &_Yso,   "_Yso/F");
    fTraTree -> Branch("SourceZ",   &_Zso,   "_Zso/F");
    fTraTree -> Branch("SourceE",   &_Eso,   "_Eso/F");
    fTraTree -> Branch("SourceTH",  &_THso,  "_THso/F");
    fTraTree -> Branch("SourcePHI", &_PHIso, "_PHIso/F");
}


void VKRecorderRoot::FillTree() 
{
    if (fTFile == NULL) return;

    // zero suppression
    G4bool zero   = fZeroSuppression && fEDepSignal==0 && fEDepVeto==0;
    G4bool zeroV2 = fZeroSuppression && fEDepV2Front==0 && fEDepV2Signal==0 && fEDepV2Veto==0;

    // beta detector
    
    _eDepSignal = fEDepSignal /keV;
    _eDepVeto   = fEDepVeto /keV;
    
    _eDepV2Front  = fEDepV2Front /keV;
    _eDepV2Signal = fEDepV2Signal /keV;
    _eDepV2Veto   = fEDepV2Veto /keV;

    _photonsSignal = fPhotonsSignal;
    _photonsVeto   = fPhotonsVeto;
    
    if ((fBetaOn && !zero) || (fBetaV2On && !zeroV2)) fDetTree -> Fill();

    // virtual detectors
    
    // focal plane
    _Xfp = fX_fp /mm;
    _Yfp = fY_fp /mm;
    _Zfp = fZ_fp /mm;
    _Efp = fEkin_fp /keV;
    _THfp = fTheta_fp /deg;
    _PHIfp = fPhi_fp /deg;

    // midway plane
    _Xmp = fX_mp /mm;
    _Ymp = fY_mp /mm;
    _Zmp = fZ_mp /mm;
    _Emp = fEkin_mp /keV;
    _THmp = fTheta_mp /deg;
    _PHImp = fPhi_mp /deg;

    // source
    _Xso = fX_so /mm;
    _Yso = fY_so /mm;
    _Zso = fZ_so /mm;
    _Eso = fEkin_so /keV;
    _THso = fTheta_so /deg;
    _PHIso = fPhi_so /deg;

    if (fVirtualOn) fTraTree -> Fill();
}


void VKRecorderRoot::ClearRootBranches() 
{
    _eDepSignal = _eDepVeto = 0;
    _eDepV2Front = _eDepV2Signal = _eDepV2Veto = 0;
    _photonsSignal = _photonsVeto = 0;

    _Xfp = _Yfp = _Zfp = _Efp = _THfp = _PHIfp = 0;
    _Xmp = _Ymp = _Zmp = _Emp = _THmp = _PHImp = 0;
    _Xso = _Yso = _Zso = _Eso = _THso = _PHIso = 0;
}


void VKRecorderRoot::OpenRootFile(G4String filename)
{
    // If TFile is already open, write data to it and close it
    if (fTFile != NULL) WriteToRootFileAndClose();

    // Set new filename
    fFilename = filename;
        
    // Open TFile
    fTFile   = new TFile(fFilename, "RECREATE");
    G4cout << " ---> Open file: " << fFilename << G4endl;

    // Create trees
    if (fBetaOn || fBetaV2On) {
        G4cout << " ---> Create Detector tree" << G4endl;
        fDetTree = new TTree("Detector", "Detector data");
        CreateDetBranches();
    }
    if (fVirtualOn) {
        G4cout << " ---> Create Trajectory tree" << G4endl;
        fTraTree = new TTree("Trajectory", "Trajectory data");
        CreateTraBranches();
    }
}


void VKRecorderRoot::WriteToRootFileAndClose()
{
    if (fTFile == NULL) return;

    G4cout << " ---> Write to file: " << fFilename << G4endl;

    if (fBetaOn || fBetaV2On) fDetTree -> Write();
    if (fVirtualOn) fTraTree -> Write();

    CloseRootFile();
}


void VKRecorderRoot::CloseRootFile()
{
    if (fTFile == NULL) return;

    G4cout << " ---> Close file: " << fFilename << G4endl;

    fTFile -> Close();
    fTFile = NULL;

	fFilename = ""; // clear file name
}


void VKRecorderRoot::WriteMetaDataToRootFile()
{
    if (fTFile == NULL) return;

    // get pointer to run manager
    G4RunManager * rm = G4RunManager::GetRunManager();

    // get pointer to primary generator action    
    const VKPrimaryGeneratorAction * pga = 
        static_cast<const VKPrimaryGeneratorAction*>(rm->GetUserPrimaryGeneratorAction());

    // Source area
    G4double area = pga -> GetSourceArea();
    TParameter<double>* pArea = new TParameter<double>("SOURCE_AREA/cm2", area/cm2);
    TNamed n1("SOURCE_AREA", Form("%f cm^2", area/cm2));
    n1.Write();

    // Sampling coverage
    double coverage = pga -> GetCoverage();
    TParameter<double>* pCoverage = new TParameter<double>("SAMPLING_COVERAGE", coverage);
    TNamed n2("SAMPLING_COVERAGE", Form("%f", coverage));
    n2.Write();
    
    // Number of events to be simulated
    int numberOfEvents = rm -> GetCurrentRun() -> GetNumberOfEventToBeProcessed();
    TParameter<int>* pEvents = new TParameter<int>("SIMULATED_EVENTS", numberOfEvents);
    TNamed n3("SIMULATED_EVENTS", Form("%i", numberOfEvents));
    n3.Write();
    
    // Normalization
    double normalization = numberOfEvents / coverage;
    TParameter<double>* pNorm = new TParameter<double>("NORMALIZATION", normalization);
    
    // Write meta data to detector tree
    if (fBetaOn || fBetaV2On) {
        fDetTree -> GetUserInfo() -> Add(pArea);
        fDetTree -> GetUserInfo() -> Add(pCoverage);
        fDetTree -> GetUserInfo() -> Add(pEvents);
        fDetTree -> GetUserInfo() -> Add(pNorm);
    }

    // Dump contents of run macro
    if (fRunMacroFileName != "") {
        string runmacCont = "\n\nRun macro: " + fRunMacroFileName + "\n";
        ifstream runmac(fRunMacroFileName);
        stringstream ss;
        ss << runmac.rdbuf();   // read the file
        runmacCont += ss.str(); // content of the file
        runmacCont += "\n";
        TNamed macro("macro", runmacCont.c_str());
        macro.Write();
    }
}
